/**
 * apeman app to handle sign.
 * @memberof module:apeman-app-sign/lib
 * @function create
 * @param {Apemandb} db - Database to work with.
 * @param {object} [options] - Optional settings.
 * @param {string} [options.userModel='User'] - Name of user model.
 * @param {string} [options.signModel='Sign'] - Name of sign model.
 * @param {function} [options.beforeSignget=noop()] - Handler before signget.
 * @param {function} [options.beforeSignup=noop()] - Handler before signup.
 * @param {function} [options.beforeSignin=noop()] - Handler before signin.
 * @param {function} [options.beforeSignout=noop()] - Handler before signout.
 * @param {function} [options.beforeSignmod=noop()] - Handler before signmod.
 * @param {function} [options.beforeSigndel=noop()] - Handler before signdel.
 * @param {object} [options.signinSchema=defaultSchema('signin')] - Schema for signin endpoint.
 * @param {object} [options.signupSchema=defaultSchema('signup')] - Schema for signup endpoint.
 * @param {object} [options.signmodSchema=defaultSchema('signmod')] - Schema for signmod endpoint.
 * @param {object} [options.signdelSchema=defaultSchema('signdel')] - Schema for signdel endpoint.
 * @param {object} [options.signdelSchema=defaultSchema('signdel')] - Schema for signdel endpoint.
 * @param {object} [options.signoutSchema=defaultSchema('signout')] - Schema for signout endpoint.
 * @param {string|string[]} [signKey=['account_name', 'email']] - Key(s) of user to use as sign key.
 * @param {string} [options.format] - Response format. "json", "xml", etc.
 * @param {number} [options.saltLength=24] - Length of salt.
 * @param {number} [options.passwordSpec='pdkdf2:120:84'] - Password algorithm, iterations, length.
 * @returns {function} - Defined app function.
 */

"use strict";

var argx = require('argx'),
    route = require('apeman-app-route'),
    validate = require('apeman-app-validate'),
    noop = require('apeman-app-noop'),
    action = require('apeman-app-action'),
    objnest = require('objnest'),
    async = require('async'),
    assert = require('assert'),
    defineDigest = require('./digesting/define_digest'),
    doSignup = require('./signing/do_signup'),
    doSignget = require('./signing/do_signget'),
    doSigndel = require('./signing/do_signdel'),
    doSignin = require('./signing/do_signin'),
    doSignmod = require('./signing/do_signmod'),
    doSignout = require('./signing/do_signout'),
    schemas = require('./schemas');

/** @lends create */
function create(db, options) {
    var args = argx(arguments);
    options = objnest.expand(args.pop('object') || {});

    var User = db.model(options.userModel || 'User'),
        Sign = db.model(options.signModel || 'Sign');

    var digest = defineDigest(options.passwordSpec || 'pdkdf2:120:84');

    var signKeys = [].concat(options.signKey || [
            'account_name', 'email'
        ]);
    var format = options.format;

    var beforeSignget = options.beforeSignget || noop(),
        beforeSignup = options.beforeSignup || noop(),
        beforeSignin = options.beforeSignin || noop(),
        beforeSignout = options.beforeSignout || noop(),
        beforeSignmod = options.beforeSignmod || noop(),
        beforeSigndel = options.beforeSigndel || noop()
        ;

    var handlers = [
        function (req, res, next) {
            next();
        },
        route({
            '/signget': {
                'GET': [
                    beforeSignget,
                    action('signget'),
                    function signget(req, res, next) {
                        var signed = req.session.signed;
                        if (!signed) {
                            res.statusCode = 404;
                            res.send({}, format);
                            return;
                        }
                        doSignget(User, Sign, signed)
                            .then(function (signed) {
                                req.signed(signed);
                                if (signed) {
                                    res.send(signed, format);
                                } else {
                                    res.statusCode = 404;
                                    res.send({}, format);
                                }
                            })
                            .catch(next);
                    }
                ]
            },
            '/signup': {
                'POST': [
                    beforeSignup,
                    validate(options.signupSchema || schemas('signup')),
                    action('signup'),
                    function signup(req, res, next) {
                        var body = req.body;
                        doSignup(User, Sign, body, digest, signKeys, options.saltLength)
                            .then(function (result) {
                                res.send({
                                    success: true,
                                    user: result.user.toJSON(),
                                    sign: result.signs.map(function (sign) {
                                        return sign.toJSON();
                                    })
                                }, format);
                            })
                            .catch(next);
                    }
                ]
            },
            '/signin': {
                'POST': [
                    beforeSignin,
                    validate(options.signinSchema || schemas('signin')),
                    action('signin'),
                    function signin(req, res, next) {
                        var body = req.body;

                        doSignin(User, Sign, body, digest)
                            .then(function (signed) {
                                req.signed(signed);
                                var success = !!signed;
                                res.statusCode = success ? 200 : 400;
                                res.send({
                                    success: success,
                                    signed: signed
                                }, format);
                            })
                            .catch(next);
                    }
                ]
            },
            '/signout': {
                'POST': [
                    beforeSignout,
                    validate(options.signoutSchema || schemas('signout')),
                    action('signout'),
                    function signout(req, res, next) {
                        var signed = req.session.signed;
                        doSignout(User, Sign, signed)
                            .then(function (signed) {
                                req.signed(null);
                                res.send({
                                    success: true,
                                    signed: signed
                                }, format);
                            })
                            .catch(next);
                    }
                ]
            },
            '/signmod': {
                'POST': [
                    beforeSignmod,
                    validate(options.signmodSchema || schemas('signmod')),
                    action('signmod'),
                    function signmod(req, res, next) {
                        var body = req.body;
                        var signed = req.session.signed;
                        doSignmod(User, Sign, body, digest, signed)
                            .then(function (signed) {
                                req.signed(signed);
                                var success = !!signed;
                                req.statusCode = signed ? 200 : 400;
                                res.send({
                                    success: success,
                                    signed: signed
                                }, format);
                            })
                            .catch(next);
                    }
                ]
            },
            '/signdel': {
                'POST': [
                    beforeSigndel,
                    validate(options.signdelSchema || schemas('signdel')),
                    action('signdel'),
                    function signdel(req, res, next) {
                        var signed = req.session.signed;
                        doSigndel(User, Sign, signed)
                            .then(function () {
                                res.send({
                                    success: true
                                }, format);
                            })
                            .catch(next);
                    }
                ]
            }
        })
    ];


    /**
     * Defined app.
     * @function app
     * @param {object} req - Request object.
     * @param {object} res - Response object.
     * @param {function} next - Pass to next handler.
     */
    function app(req, res, next) {
        assert(req.signed, 'req.signed not found. Use "apeman-app-signed" before this.');
        assert(req.body, 'req.body not found. Use "apeman-app-form" before this.');
        assert(res.header, 'req.header not found. Use "apeman-app-header" before this.');
        assert(res.send, 'req.send not found. Use "apeman-app-send" before this.');

        app.$handler(req, res, next);
    }

    app.$handler = function (req, res, next) {
        async.eachSeries(handlers, function (handler, next) {
            handler(req, res, next);
        }, next);
    };

    app.$endpoints = handlers[1].$endpoints;


    // Description of this app.
    app.$desc = "Sign endpoint.";
    return app;
}

module.exports = create;
